{
 "cells": [
  {
   "cell_type": "markdown",
   "source": [
    "Surname Classification with RNN\r\n",
    "===============================\r\n",
    "\r\n",
    "In this example, we see surname classification in which character sequences (surnames) are classified to nationality of origin. \r\n",
    "\r\n",
    "Inferring demographic information (like nationality) from publicly observable data has applications from product recommendations to ensuring fair outcomes for users across different demographics. However, demographic and other self­-identifying attributes are collectively called \"protected attributes\". We must exercise care in the use of such attributes in modeling and in products. We begin by splitting the characters of each surname and treating them the same way we treated words. Aside from the data difference, character­ level models are mostly similar to word­based models in structure and implementation. "
   ],
   "metadata": {}
  },
  {
   "cell_type": "code",
   "execution_count": 13,
   "source": [
    "from argparse import Namespace\r\n",
    "import os\r\n",
    "import json\r\n",
    "\r\n",
    "import numpy as np\r\n",
    "import pandas as pd\r\n",
    "import torch\r\n",
    "import torch.nn as nn\r\n",
    "import torch.nn.functional as F\r\n",
    "import torch.optim as optim\r\n",
    "from torch.utils.data import Dataset, DataLoader\r\n",
    "from tqdm.notebook import tqdm"
   ],
   "outputs": [],
   "metadata": {}
  },
  {
   "cell_type": "markdown",
   "source": [
    "## Vocabulary, Vectorizer, Dataset\r\n",
    "\r\n",
    "We treat each surname as a sequence of characters. As usual, we implement a dataset class, that returns the vectorized surname as well as the integer representing its nationality. Additionally returned is the length of the sequence, which is used in downstream computations to know where the final vector in the sequence is located. This is a part of the familiar sequence of steps - implementing Dataset, a Vectorizer, and a Vocabulary before the actual training can take place."
   ],
   "metadata": {}
  },
  {
   "cell_type": "code",
   "execution_count": 14,
   "source": [
    "class Vocabulary(object):\r\n",
    "    \"\"\"Class to process text and extract vocabulary for mapping\"\"\"\r\n",
    "\r\n",
    "    def __init__(self, token_to_idx=None):\r\n",
    "        \"\"\"\r\n",
    "        Args:\r\n",
    "            token_to_idx (dict): a pre-existing map of tokens to indices\r\n",
    "        \"\"\"\r\n",
    "\r\n",
    "        if token_to_idx is None:\r\n",
    "            token_to_idx = {}\r\n",
    "        self._token_to_idx = token_to_idx\r\n",
    "\r\n",
    "        self._idx_to_token = {idx: token \r\n",
    "                              for token, idx in self._token_to_idx.items()}\r\n",
    "        \r\n",
    "    def to_serializable(self):\r\n",
    "        \"\"\" returns a dictionary that can be serialized \"\"\"\r\n",
    "        return {'token_to_idx': self._token_to_idx}\r\n",
    "\r\n",
    "    @classmethod\r\n",
    "    def from_serializable(cls, contents):\r\n",
    "        \"\"\" instantiates the Vocabulary from a serialized dictionary \"\"\"\r\n",
    "        return cls(**contents)\r\n",
    "\r\n",
    "    def add_token(self, token):\r\n",
    "        \"\"\"Update mapping dicts based on the token.\r\n",
    "\r\n",
    "        Args:\r\n",
    "            token (str): the item to add into the Vocabulary\r\n",
    "        Returns:\r\n",
    "            index (int): the integer corresponding to the token\r\n",
    "        \"\"\"\r\n",
    "        if token in self._token_to_idx:\r\n",
    "            index = self._token_to_idx[token]\r\n",
    "        else:\r\n",
    "            index = len(self._token_to_idx)\r\n",
    "            self._token_to_idx[token] = index\r\n",
    "            self._idx_to_token[index] = token\r\n",
    "        return index\r\n",
    "            \r\n",
    "    def add_many(self, tokens):\r\n",
    "        \"\"\"Add a list of tokens into the Vocabulary\r\n",
    "        \r\n",
    "        Args:\r\n",
    "            tokens (list): a list of string tokens\r\n",
    "        Returns:\r\n",
    "            indices (list): a list of indices corresponding to the tokens\r\n",
    "        \"\"\"\r\n",
    "        return [self.add_token(token) for token in tokens]\r\n",
    "\r\n",
    "    def lookup_token(self, token):\r\n",
    "        \"\"\"Retrieve the index associated with the token \r\n",
    "        \r\n",
    "        Args:\r\n",
    "            token (str): the token to look up \r\n",
    "        Returns:\r\n",
    "            index (int): the index corresponding to the token\r\n",
    "        \"\"\"\r\n",
    "        return self._token_to_idx[token]\r\n",
    "\r\n",
    "    def lookup_index(self, index):\r\n",
    "        \"\"\"Return the token associated with the index\r\n",
    "        \r\n",
    "        Args: \r\n",
    "            index (int): the index to look up\r\n",
    "        Returns:\r\n",
    "            token (str): the token corresponding to the index\r\n",
    "        Raises:\r\n",
    "            KeyError: if the index is not in the Vocabulary\r\n",
    "        \"\"\"\r\n",
    "        if index not in self._idx_to_token:\r\n",
    "            raise KeyError(\"the index (%d) is not in the Vocabulary\" % index)\r\n",
    "        return self._idx_to_token[index]\r\n",
    "\r\n",
    "    def __str__(self):\r\n",
    "        return \"<Vocabulary(size=%d)>\" % len(self)\r\n",
    "\r\n",
    "    def __len__(self):\r\n",
    "        return len(self._token_to_idx)"
   ],
   "outputs": [],
   "metadata": {}
  },
  {
   "cell_type": "markdown",
   "source": [
    "The first stage in the vectorization pipeline is to map each character token in the surname to a unique integer. To accomplish this, we use the `SequenceVocabulary` data structure, which we first introduced and described in [Section 5.2.2 of Document Classification using CNN and GloVe Embeddings](https://weiliu2k.github.io/CITS4012/cnn/Document_Classification_with_CNN.html#data-vectorization-classes). Recall that this data structure not only maps characters in the names to integers, but also utilizes four special­purpose tokens: \r\n",
    "- the `UNK` token, \r\n",
    "- the `MASK` token, \r\n",
    "- the BEGIN­OF­SEQUENCE token, and \r\n",
    "- the END­OF­SEQUENCE token. \r\n",
    "\r\n",
    "The first two tokens are vital for language data: the `UNK` token is used for unseen Out­Of­Vocabulary (OOV) tokens in the input and the `MASK` token enables handling variable­-length inputs. The latter two tokens provide the model with sentence boundary features and are prepended and appended to the sequence, respectively. "
   ],
   "metadata": {}
  },
  {
   "cell_type": "code",
   "execution_count": 15,
   "source": [
    "class SequenceVocabulary(Vocabulary):\r\n",
    "    def __init__(self, token_to_idx=None, unk_token=\"<UNK>\",\r\n",
    "                 mask_token=\"<MASK>\", begin_seq_token=\"<BEGIN>\",\r\n",
    "                 end_seq_token=\"<END>\"):\r\n",
    "\r\n",
    "        super(SequenceVocabulary, self).__init__(token_to_idx)\r\n",
    "\r\n",
    "        self._mask_token = mask_token\r\n",
    "        self._unk_token = unk_token\r\n",
    "        self._begin_seq_token = begin_seq_token\r\n",
    "        self._end_seq_token = end_seq_token\r\n",
    "\r\n",
    "        self.mask_index = self.add_token(self._mask_token)\r\n",
    "        self.unk_index = self.add_token(self._unk_token)\r\n",
    "        self.begin_seq_index = self.add_token(self._begin_seq_token)\r\n",
    "        self.end_seq_index = self.add_token(self._end_seq_token)\r\n",
    "\r\n",
    "    def to_serializable(self):\r\n",
    "        contents = super(SequenceVocabulary, self).to_serializable()\r\n",
    "        contents.update({'unk_token': self._unk_token,\r\n",
    "                         'mask_token': self._mask_token,\r\n",
    "                         'begin_seq_token': self._begin_seq_token,\r\n",
    "                         'end_seq_token': self._end_seq_token})\r\n",
    "        return contents\r\n",
    "\r\n",
    "    def lookup_token(self, token):\r\n",
    "        \"\"\"Retrieve the index associated with the token \r\n",
    "          or the UNK index if token isn't present.\r\n",
    "        \r\n",
    "        Args:\r\n",
    "            token (str): the token to look up \r\n",
    "        Returns:\r\n",
    "            index (int): the index corresponding to the token\r\n",
    "        Notes:\r\n",
    "            `unk_index` needs to be >=0 (having been added into the Vocabulary) \r\n",
    "              for the UNK functionality \r\n",
    "        \"\"\"\r\n",
    "        if self.unk_index >= 0:\r\n",
    "            return self._token_to_idx.get(token, self.unk_index)\r\n",
    "        else:\r\n",
    "            return self._token_to_idx[token]"
   ],
   "outputs": [],
   "metadata": {}
  },
  {
   "cell_type": "markdown",
   "source": [
    "The overall vectorizer is `SurnameVectorizer`, which populates the `SequenceVocabulary` by surname characters, and the normal `Vocabulary` for nationalities. "
   ],
   "metadata": {}
  },
  {
   "cell_type": "code",
   "execution_count": 16,
   "source": [
    "class SurnameVectorizer(object):\r\n",
    "    \"\"\" The Vectorizer which coordinates the Vocabularies and puts them to use\"\"\"   \r\n",
    "    def __init__(self, char_vocab, nationality_vocab):\r\n",
    "        \"\"\"\r\n",
    "        Args:\r\n",
    "            char_vocab (Vocabulary): maps characters to integers\r\n",
    "            nationality_vocab (Vocabulary): maps nationalities to integers\r\n",
    "        \"\"\"\r\n",
    "        self.char_vocab = char_vocab\r\n",
    "        self.nationality_vocab = nationality_vocab\r\n",
    "\r\n",
    "    def vectorize(self, surname, vector_length=-1):\r\n",
    "        \"\"\"\r\n",
    "        Args:\r\n",
    "            title (str): the string of characters\r\n",
    "            vector_length (int): an argument for forcing the length of index vector\r\n",
    "        \"\"\"\r\n",
    "        indices = [self.char_vocab.begin_seq_index]\r\n",
    "        indices.extend(self.char_vocab.lookup_token(token) \r\n",
    "                       for token in surname)\r\n",
    "        indices.append(self.char_vocab.end_seq_index)\r\n",
    "\r\n",
    "        if vector_length < 0:\r\n",
    "            vector_length = len(indices)\r\n",
    "\r\n",
    "        out_vector = np.zeros(vector_length, dtype=np.int64)         \r\n",
    "        out_vector[:len(indices)] = indices\r\n",
    "        out_vector[len(indices):] = self.char_vocab.mask_index\r\n",
    "        \r\n",
    "        return out_vector, len(indices)\r\n",
    "\r\n",
    "    @classmethod\r\n",
    "    def from_dataframe(cls, surname_df):\r\n",
    "        \"\"\"Instantiate the vectorizer from the dataset dataframe\r\n",
    "        \r\n",
    "        Args:\r\n",
    "            surname_df (pandas.DataFrame): the surnames dataset\r\n",
    "        Returns:\r\n",
    "            an instance of the SurnameVectorizer\r\n",
    "        \"\"\"\r\n",
    "        char_vocab = SequenceVocabulary()\r\n",
    "        nationality_vocab = Vocabulary()\r\n",
    "\r\n",
    "        for index, row in surname_df.iterrows():\r\n",
    "            for char in row.surname:\r\n",
    "                char_vocab.add_token(char)\r\n",
    "            nationality_vocab.add_token(row.nationality)\r\n",
    "\r\n",
    "        return cls(char_vocab, nationality_vocab)\r\n",
    "\r\n",
    "    @classmethod\r\n",
    "    def from_serializable(cls, contents):\r\n",
    "        char_vocab = SequenceVocabulary.from_serializable(contents['char_vocab'])\r\n",
    "        nat_vocab =  Vocabulary.from_serializable(contents['nationality_vocab'])\r\n",
    "\r\n",
    "        return cls(char_vocab=char_vocab, nationality_vocab=nat_vocab)\r\n",
    "\r\n",
    "    def to_serializable(self):\r\n",
    "        return {'char_vocab': self.char_vocab.to_serializable(), \r\n",
    "                'nationality_vocab': self.nationality_vocab.to_serializable()}"
   ],
   "outputs": [],
   "metadata": {}
  },
  {
   "cell_type": "code",
   "execution_count": 17,
   "source": [
    "class SurnameDataset(Dataset):\r\n",
    "    def __init__(self, surname_df, vectorizer):\r\n",
    "        \"\"\"\r\n",
    "        Args:\r\n",
    "            surname_df (pandas.DataFrame): the dataset\r\n",
    "            vectorizer (SurnameVectorizer): vectorizer instatiated from dataset\r\n",
    "        \"\"\"\r\n",
    "        self.surname_df = surname_df \r\n",
    "        self._vectorizer = vectorizer\r\n",
    "\r\n",
    "        self._max_seq_length = max(map(len, self.surname_df.surname)) + 2\r\n",
    "\r\n",
    "        self.train_df = self.surname_df[self.surname_df.split=='train']\r\n",
    "        self.train_size = len(self.train_df)\r\n",
    "\r\n",
    "        self.val_df = self.surname_df[self.surname_df.split=='val']\r\n",
    "        self.validation_size = len(self.val_df)\r\n",
    "\r\n",
    "        self.test_df = self.surname_df[self.surname_df.split=='test']\r\n",
    "        self.test_size = len(self.test_df)\r\n",
    "\r\n",
    "        self._lookup_dict = {'train': (self.train_df, self.train_size), \r\n",
    "                             'val': (self.val_df, self.validation_size), \r\n",
    "                             'test': (self.test_df, self.test_size)}\r\n",
    "\r\n",
    "        self.set_split('train')\r\n",
    "        \r\n",
    "        # Class weights\r\n",
    "        class_counts = self.train_df.nationality.value_counts().to_dict()\r\n",
    "        def sort_key(item):\r\n",
    "            return self._vectorizer.nationality_vocab.lookup_token(item[0])\r\n",
    "        sorted_counts = sorted(class_counts.items(), key=sort_key)\r\n",
    "        frequencies = [count for _, count in sorted_counts]\r\n",
    "        self.class_weights = 1.0 / torch.tensor(frequencies, dtype=torch.float32)\r\n",
    "\r\n",
    "        \r\n",
    "    @classmethod\r\n",
    "    def load_dataset_and_make_vectorizer(cls, surname_csv):\r\n",
    "        \"\"\"Load dataset and make a new vectorizer from scratch\r\n",
    "        \r\n",
    "        Args:\r\n",
    "            surname_csv (str): location of the dataset\r\n",
    "        Returns:\r\n",
    "            an instance of SurnameDataset\r\n",
    "        \"\"\"\r\n",
    "        surname_df = pd.read_csv(surname_csv)\r\n",
    "        train_surname_df = surname_df[surname_df.split=='train']\r\n",
    "        return cls(surname_df, SurnameVectorizer.from_dataframe(train_surname_df))\r\n",
    "        \r\n",
    "    @classmethod\r\n",
    "    def load_dataset_and_load_vectorizer(cls, surname_csv, vectorizer_filepath):\r\n",
    "        \"\"\"Load dataset and the corresponding vectorizer. \r\n",
    "        Used in the case in the vectorizer has been cached for re-use\r\n",
    "        \r\n",
    "        Args:\r\n",
    "            surname_csv (str): location of the dataset\r\n",
    "            vectorizer_filepath (str): location of the saved vectorizer\r\n",
    "        Returns:\r\n",
    "            an instance of SurnameDataset\r\n",
    "        \"\"\"\r\n",
    "        surname_df = pd.read_csv(surname_csv)\r\n",
    "        vectorizer = cls.load_vectorizer_only(vectorizer_filepath)\r\n",
    "        return cls(surname_df, vectorizer)\r\n",
    "\r\n",
    "    @staticmethod\r\n",
    "    def load_vectorizer_only(vectorizer_filepath):\r\n",
    "        \"\"\"a static method for loading the vectorizer from file\r\n",
    "        \r\n",
    "        Args:\r\n",
    "            vectorizer_filepath (str): the location of the serialized vectorizer\r\n",
    "        Returns:\r\n",
    "            an instance of SurnameVectorizer\r\n",
    "        \"\"\"\r\n",
    "        with open(vectorizer_filepath) as fp:\r\n",
    "            return SurnameVectorizer.from_serializable(json.load(fp))\r\n",
    "\r\n",
    "    def save_vectorizer(self, vectorizer_filepath):\r\n",
    "        \"\"\"saves the vectorizer to disk using json\r\n",
    "        \r\n",
    "        Args:\r\n",
    "            vectorizer_filepath (str): the location to save the vectorizer\r\n",
    "        \"\"\"\r\n",
    "        with open(vectorizer_filepath, \"w\") as fp:\r\n",
    "            json.dump(self._vectorizer.to_serializable(), fp)\r\n",
    "\r\n",
    "    def get_vectorizer(self):\r\n",
    "        \"\"\" returns the vectorizer \"\"\"\r\n",
    "        return self._vectorizer\r\n",
    "\r\n",
    "    def set_split(self, split=\"train\"):\r\n",
    "        self._target_split = split\r\n",
    "        self._target_df, self._target_size = self._lookup_dict[split]\r\n",
    "\r\n",
    "    def __len__(self):\r\n",
    "        return self._target_size\r\n",
    "\r\n",
    "    def __getitem__(self, index):\r\n",
    "        \"\"\"the primary entry point method for PyTorch datasets\r\n",
    "        \r\n",
    "        Args:\r\n",
    "            index (int): the index to the data point \r\n",
    "        Returns:\r\n",
    "            a dictionary holding the data point's:\r\n",
    "                features (x_data)\r\n",
    "                label (y_target)\r\n",
    "                feature length (x_length)\r\n",
    "        \"\"\"\r\n",
    "        row = self._target_df.iloc[index]\r\n",
    "        \r\n",
    "        surname_vector, vec_length = \\\r\n",
    "            self._vectorizer.vectorize(row.surname, self._max_seq_length)\r\n",
    "        \r\n",
    "        nationality_index = \\\r\n",
    "            self._vectorizer.nationality_vocab.lookup_token(row.nationality)\r\n",
    "\r\n",
    "        return {'x_data': surname_vector, \r\n",
    "                'y_target': nationality_index, \r\n",
    "                'x_length': vec_length}\r\n",
    "\r\n",
    "    def get_num_batches(self, batch_size):\r\n",
    "        \"\"\"Given a batch size, return the number of batches in the dataset\r\n",
    "        \r\n",
    "        Args:\r\n",
    "            batch_size (int)\r\n",
    "        Returns:\r\n",
    "            number of batches in the dataset\r\n",
    "        \"\"\"\r\n",
    "        return len(self) // batch_size\r\n",
    "\r\n",
    "    \r\n",
    "\r\n",
    "def generate_batches(dataset, batch_size, shuffle=True,\r\n",
    "                     drop_last=True, device=\"cpu\"): \r\n",
    "    \"\"\"\r\n",
    "    A generator function which wraps the PyTorch DataLoader. It will \r\n",
    "      ensure each tensor is on the write device location.\r\n",
    "    \"\"\"\r\n",
    "    dataloader = DataLoader(dataset=dataset, batch_size=batch_size,\r\n",
    "                            shuffle=shuffle, drop_last=drop_last)\r\n",
    "\r\n",
    "    for data_dict in dataloader:\r\n",
    "        out_data_dict = {}\r\n",
    "        for name, tensor in data_dict.items():\r\n",
    "            out_data_dict[name] = data_dict[name].to(device)\r\n",
    "        yield out_data_dict"
   ],
   "outputs": [],
   "metadata": {}
  },
  {
   "cell_type": "markdown",
   "source": [
    "## Model\r\n",
    "\r\n",
    "The `SurnameClassifier` model is composed of an Embedding layer, the ElmanRNN, and a Linear layer. We assume that the input to the model is tokens represented as a set of integers after they have been mapped to integers by the `SequenceVocabulary`. The model first embeds the integers using the embedding layer. Then, using the RNN, sequence representation vectors are computed. These vectors represent the hidden state for each character in the surname. Because the goal is to classify each surname,** the vector corresponding to the final character position in each surname is extracted**. One way to think about this vector is that the final vector is a result of passing over the entire sequence input, and hence it's a summary vector for the surname. These summary vectors are passed through the Linear layer to compute a prediction vector. The prediction vector is used in the training loss, or we can apply the softmax function to create a probability distribution over surnames.\r\n",
    "\r\n",
    "The arguments to the model are \r\n",
    "\r\n",
    "- the size of the embeddings - *hyperparameter*,\r\n",
    "- the number of embeddings (i.e., vocabulary size) - determined by the data,\r\n",
    "- the number of classes - determined by the data, and \r\n",
    "- the hidden state size of the RNN - *hyperparameter*.\r\n",
    "\r\n",
    ":::{important}\r\n",
    "Although the two hyperparameters can take on any value, it is usually good to start with something small that ensures fast training, as a way verifying that the model actually works. \r\n",
    ":::"
   ],
   "metadata": {}
  },
  {
   "cell_type": "markdown",
   "source": [
    "### Retrieving the last vector of each sequence\r\n",
    "\r\n",
    "You will notice that the `forward()` function in the model requires the lengths of the sequences. The lengths are used to retrieve the final vector of each sequence in the tensor that is returned from the RNN with a function named `column_gather()`, shown below. The function iterates over batch row indices and retrieves the vector that's at the position indicated by the\r\n",
    "corresponding length of sequence."
   ],
   "metadata": {}
  },
  {
   "cell_type": "code",
   "execution_count": 18,
   "source": [
    "\r\n",
    "def column_gather(y_out, x_lengths):\r\n",
    "    '''Get a specific vector from each batch datapoint in `y_out`.\r\n",
    "\r\n",
    "    More precisely, iterate over batch row indices, get the vector that's at\r\n",
    "    the position indicated by the corresponding value in `x_lengths` at the row\r\n",
    "    index.\r\n",
    "\r\n",
    "    Args:\r\n",
    "        y_out (torch.FloatTensor, torch.cuda.FloatTensor)\r\n",
    "            shape: (batch, sequence, feature)\r\n",
    "        x_lengths (torch.LongTensor, torch.cuda.LongTensor)\r\n",
    "            shape: (batch,)\r\n",
    "\r\n",
    "    Returns:\r\n",
    "        y_out (torch.FloatTensor, torch.cuda.FloatTensor)\r\n",
    "            shape: (batch, feature)\r\n",
    "    '''\r\n",
    "    x_lengths = x_lengths.long().detach().cpu().numpy() - 1\r\n",
    "\r\n",
    "    out = []\r\n",
    "    for batch_index, column_index in enumerate(x_lengths):\r\n",
    "        out.append(y_out[batch_index, column_index])\r\n",
    "\r\n",
    "    return torch.stack(out)\r\n"
   ],
   "outputs": [],
   "metadata": {}
  },
  {
   "cell_type": "markdown",
   "source": [
    "### The ElmanRNN model"
   ],
   "metadata": {}
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "source": [
    "\r\n",
    "class ElmanRNN(nn.Module):\r\n",
    "    \"\"\" an Elman RNN built using the RNNCell \"\"\"\r\n",
    "    def __init__(self, input_size, hidden_size, batch_first=False):\r\n",
    "        \"\"\"\r\n",
    "        Args:\r\n",
    "            input_size (int): size of the input vectors\r\n",
    "            hidden_size (int): size of the hidden state vectors\r\n",
    "            bathc_first (bool): whether the 0th dimension is batch\r\n",
    "        \"\"\"\r\n",
    "        super(ElmanRNN, self).__init__()\r\n",
    "        \r\n",
    "        self.rnn_cell = nn.RNNCell(input_size, hidden_size)\r\n",
    "        \r\n",
    "        self.batch_first = batch_first\r\n",
    "        self.hidden_size = hidden_size\r\n",
    "\r\n",
    "    def _initial_hidden(self, batch_size):\r\n",
    "        return torch.zeros((batch_size, self.hidden_size))\r\n",
    "\r\n",
    "    def forward(self, x_in, initial_hidden=None):\r\n",
    "        \"\"\"The forward pass of the ElmanRNN\r\n",
    "        \r\n",
    "        Args:\r\n",
    "            x_in (torch.Tensor): an input data tensor. \r\n",
    "                If self.batch_first: x_in.shape = (batch, seq_size, feat_size)\r\n",
    "                Else: x_in.shape = (seq_size, batch, feat_size)\r\n",
    "            initial_hidden (torch.Tensor): the initial hidden state for the RNN\r\n",
    "        Returns:\r\n",
    "            hiddens (torch.Tensor): The outputs of the RNN at each time step. \r\n",
    "                If self.batch_first: hiddens.shape = (batch, seq_size, hidden_size)\r\n",
    "                Else: hiddens.shape = (seq_size, batch, hidden_size)\r\n",
    "        \"\"\"\r\n",
    "        if self.batch_first:\r\n",
    "            batch_size, seq_size, feat_size = x_in.size()\r\n",
    "            x_in = x_in.permute(1, 0, 2)\r\n",
    "        else:\r\n",
    "            seq_size, batch_size, feat_size = x_in.size()\r\n",
    "    \r\n",
    "        hiddens = []\r\n",
    "\r\n",
    "        if initial_hidden is None:\r\n",
    "            initial_hidden = self._initial_hidden(batch_size)\r\n",
    "            initial_hidden = initial_hidden.to(x_in.device)\r\n",
    "\r\n",
    "        hidden_t = initial_hidden\r\n",
    "                    \r\n",
    "        for t in range(seq_size):\r\n",
    "            hidden_t = self.rnn_cell(x_in[t], hidden_t)\r\n",
    "            hiddens.append(hidden_t)\r\n",
    "            \r\n",
    "        hiddens = torch.stack(hiddens)\r\n",
    "\r\n",
    "        if self.batch_first:\r\n",
    "            hiddens = hiddens.permute(1, 0, 2)\r\n",
    "\r\n",
    "        return hiddens\r\n",
    "\r\n",
    "\r\n",
    "\r\n",
    "class SurnameClassifier(nn.Module):\r\n",
    "    \"\"\" A Classifier with an RNN to extract features and an MLP to classify \"\"\"\r\n",
    "    def __init__(self, embedding_size, num_embeddings, num_classes,\r\n",
    "                 rnn_hidden_size, batch_first=True, padding_idx=0):\r\n",
    "        \"\"\"\r\n",
    "        Args:\r\n",
    "            embedding_size (int): The size of the character embeddings\r\n",
    "            num_embeddings (int): The number of characters to embed\r\n",
    "            num_classes (int): The size of the prediction vector \r\n",
    "                Note: the number of nationalities\r\n",
    "            rnn_hidden_size (int): The size of the RNN's hidden state\r\n",
    "            batch_first (bool): Informs whether the input tensors will \r\n",
    "                have batch or the sequence on the 0th dimension\r\n",
    "            padding_idx (int): The index for the tensor padding; \r\n",
    "                see torch.nn.Embedding\r\n",
    "        \"\"\"\r\n",
    "        super(SurnameClassifier, self).__init__()\r\n",
    "\r\n",
    "        self.emb = nn.Embedding(num_embeddings=num_embeddings,\r\n",
    "                                embedding_dim=embedding_size,\r\n",
    "                                padding_idx=padding_idx)\r\n",
    "        self.rnn = ElmanRNN(input_size=embedding_size,\r\n",
    "                             hidden_size=rnn_hidden_size,\r\n",
    "                             batch_first=batch_first)\r\n",
    "        self.fc1 = nn.Linear(in_features=rnn_hidden_size,\r\n",
    "                         out_features=rnn_hidden_size)\r\n",
    "        self.fc2 = nn.Linear(in_features=rnn_hidden_size,\r\n",
    "                          out_features=num_classes)\r\n",
    "\r\n",
    "    def forward(self, x_in, x_lengths=None, apply_softmax=False):\r\n",
    "        \"\"\"The forward pass of the classifier\r\n",
    "        \r\n",
    "        Args:\r\n",
    "            x_in (torch.Tensor): an input data tensor. \r\n",
    "                x_in.shape should be (batch, input_dim)\r\n",
    "            x_lengths (torch.Tensor): the lengths of each sequence in the batch.\r\n",
    "                They are used to find the final vector of each sequence\r\n",
    "            apply_softmax (bool): a flag for the softmax activation\r\n",
    "                should be false if used with the Cross Entropy losses\r\n",
    "        Returns:\r\n",
    "            the resulting tensor. tensor.shape should be (batch, output_dim)\r\n",
    "        \"\"\"\r\n",
    "        x_embedded = self.emb(x_in)\r\n",
    "        y_out = self.rnn(x_embedded)\r\n",
    "\r\n",
    "        if x_lengths is not None:\r\n",
    "            y_out = column_gather(y_out, x_lengths)\r\n",
    "        else:\r\n",
    "            y_out = y_out[:, -1, :]\r\n",
    "\r\n",
    "        y_out = F.relu(self.fc1(F.dropout(y_out, 0.5)))\r\n",
    "        y_out = self.fc2(F.dropout(y_out, 0.5))\r\n",
    "\r\n",
    "        if apply_softmax:\r\n",
    "            y_out = F.softmax(y_out, dim=1)\r\n",
    "\r\n",
    "        return y_out"
   ],
   "outputs": [],
   "metadata": {}
  },
  {
   "cell_type": "code",
   "execution_count": 19,
   "source": [
    "def set_seed_everywhere(seed, cuda):\r\n",
    "    np.random.seed(seed)\r\n",
    "    torch.manual_seed(seed)\r\n",
    "    if cuda:\r\n",
    "        torch.cuda.manual_seed_all(seed)\r\n",
    "\r\n",
    "def handle_dirs(dirpath):\r\n",
    "    if not os.path.exists(dirpath):\r\n",
    "        os.makedirs(dirpath)"
   ],
   "outputs": [],
   "metadata": {}
  },
  {
   "cell_type": "markdown",
   "source": [
    "## Settings"
   ],
   "metadata": {}
  },
  {
   "cell_type": "code",
   "execution_count": 20,
   "source": [
    "args = Namespace(\r\n",
    "    # Data and path information\r\n",
    "    surname_csv=\"../data/surnames/surnames_with_splits.csv\",\r\n",
    "    vectorizer_file=\"vectorizer.json\",\r\n",
    "    model_state_file=\"model.pth\",\r\n",
    "    save_dir=\"model_storage/ch6/surname_classification\",\r\n",
    "    # Model hyper parameter\r\n",
    "    char_embedding_size=100,\r\n",
    "    rnn_hidden_size=64,\r\n",
    "    # Training hyper parameter\r\n",
    "    num_epochs=100,\r\n",
    "    learning_rate=1e-3,\r\n",
    "    batch_size=64,\r\n",
    "    seed=1337,\r\n",
    "    early_stopping_criteria=5,\r\n",
    "    # Runtime hyper parameter\r\n",
    "    cuda=True,\r\n",
    "    catch_keyboard_interrupt=True,\r\n",
    "    reload_from_files=False,\r\n",
    "    expand_filepaths_to_save_dir=True,\r\n",
    ")\r\n",
    "\r\n",
    "# Check CUDA\r\n",
    "if not torch.cuda.is_available():\r\n",
    "    args.cuda = False\r\n",
    "\r\n",
    "args.device = torch.device(\"cuda\" if args.cuda else \"cpu\")\r\n",
    "    \r\n",
    "print(\"Using CUDA: {}\".format(args.cuda))\r\n",
    "\r\n",
    "\r\n",
    "if args.expand_filepaths_to_save_dir:\r\n",
    "    args.vectorizer_file = os.path.join(args.save_dir,\r\n",
    "                                        args.vectorizer_file)\r\n",
    "\r\n",
    "    args.model_state_file = os.path.join(args.save_dir,\r\n",
    "                                         args.model_state_file)\r\n",
    "    \r\n",
    "# Set seed for reproducibility\r\n",
    "set_seed_everywhere(args.seed, args.cuda)\r\n",
    "\r\n",
    "# handle dirs\r\n",
    "handle_dirs(args.save_dir)"
   ],
   "outputs": [
    {
     "output_type": "stream",
     "name": "stdout",
     "text": [
      "Using CUDA: True\n"
     ]
    }
   ],
   "metadata": {
    "code_folding": []
   }
  },
  {
   "cell_type": "code",
   "execution_count": 21,
   "source": [
    "if args.reload_from_files and os.path.exists(args.vectorizer_file):\r\n",
    "    # training from a checkpoint\r\n",
    "    dataset = SurnameDataset.load_dataset_and_load_vectorizer(args.surname_csv, \r\n",
    "                                                              args.vectorizer_file)\r\n",
    "else:\r\n",
    "    # create dataset and vectorizer\r\n",
    "    dataset = SurnameDataset.load_dataset_and_make_vectorizer(args.surname_csv)\r\n",
    "    dataset.save_vectorizer(args.vectorizer_file)\r\n",
    "\r\n",
    "vectorizer = dataset.get_vectorizer()\r\n",
    "\r\n",
    "classifier = SurnameClassifier(embedding_size=args.char_embedding_size, \r\n",
    "                               num_embeddings=len(vectorizer.char_vocab),\r\n",
    "                               num_classes=len(vectorizer.nationality_vocab),\r\n",
    "                               rnn_hidden_size=args.rnn_hidden_size,\r\n",
    "                               padding_idx=vectorizer.char_vocab.mask_index)"
   ],
   "outputs": [],
   "metadata": {
    "code_folding": [
     0,
     4
    ]
   }
  },
  {
   "cell_type": "markdown",
   "source": [
    "## Training Routine\r\n",
    "\r\n",
    "The training routine follows the standard formula. For a single batch of data, apply the model and compute the prediction vectors. Use the `CrossEntropyLoss()` loss function and the ground truth to compute a loss value. Using the loss value and an optimizer, compute the gradients and update the weights of the model using those gradients. Repeat this for each batch in the training data. Proceed similarly with the validation data, but set the model in eval mode so as to prevent backpropagating. Instead, the validation data is used only to give a less­biased sense of how the model is performing. Repeat this routine for a specific number of epochs or a stopping condition (e.g. loss is less than a threshold, or loss stoped changing in the most recent two epochs) is met."
   ],
   "metadata": {}
  },
  {
   "cell_type": "code",
   "execution_count": 22,
   "source": [
    "def make_train_state(args):\r\n",
    "    return {'stop_early': False,\r\n",
    "            'early_stopping_step': 0,\r\n",
    "            'early_stopping_best_val': 1e8,\r\n",
    "            'learning_rate': args.learning_rate,\r\n",
    "            'epoch_index': 0,\r\n",
    "            'train_loss': [],\r\n",
    "            'train_acc': [],\r\n",
    "            'val_loss': [],\r\n",
    "            'val_acc': [],\r\n",
    "            'test_loss': -1,\r\n",
    "            'test_acc': -1,\r\n",
    "            'model_filename': args.model_state_file}\r\n",
    "\r\n",
    "\r\n",
    "def update_train_state(args, model, train_state):\r\n",
    "    \"\"\"Handle the training state updates.\r\n",
    "\r\n",
    "    Components:\r\n",
    "     - Early Stopping: Prevent overfitting.\r\n",
    "     - Model Checkpoint: Model is saved if the model is better\r\n",
    "    \r\n",
    "    :param args: main arguments\r\n",
    "    :param model: model to train\r\n",
    "    :param train_state: a dictionary representing the training state values\r\n",
    "    :returns:\r\n",
    "        a new train_state\r\n",
    "    \"\"\"\r\n",
    "\r\n",
    "    # Save one model at least\r\n",
    "    if train_state['epoch_index'] == 0:\r\n",
    "        torch.save(model.state_dict(), train_state['model_filename'])\r\n",
    "        train_state['stop_early'] = False\r\n",
    "\r\n",
    "    # Save model if performance improved\r\n",
    "    elif train_state['epoch_index'] >= 1:\r\n",
    "        loss_tm1, loss_t = train_state['val_loss'][-2:]\r\n",
    "         \r\n",
    "        # If loss worsened\r\n",
    "        if loss_t >= loss_tm1:\r\n",
    "            # Update step\r\n",
    "            train_state['early_stopping_step'] += 1\r\n",
    "        # Loss decreased\r\n",
    "        else:\r\n",
    "            # Save the best model\r\n",
    "            if loss_t < train_state['early_stopping_best_val']:\r\n",
    "                torch.save(model.state_dict(), train_state['model_filename'])\r\n",
    "                train_state['early_stopping_best_val'] = loss_t\r\n",
    "\r\n",
    "            # Reset early stopping step\r\n",
    "            train_state['early_stopping_step'] = 0\r\n",
    "\r\n",
    "        # Stop early ?\r\n",
    "        train_state['stop_early'] = \\\r\n",
    "            train_state['early_stopping_step'] >= args.early_stopping_criteria\r\n",
    "\r\n",
    "    return train_state\r\n",
    "\r\n",
    "\r\n",
    "def compute_accuracy(y_pred, y_target):\r\n",
    "    _, y_pred_indices = y_pred.max(dim=1)\r\n",
    "    n_correct = torch.eq(y_pred_indices, y_target).sum().item()\r\n",
    "    return n_correct / len(y_pred_indices) * 100"
   ],
   "outputs": [],
   "metadata": {
    "code_folding": []
   }
  },
  {
   "cell_type": "code",
   "execution_count": 23,
   "source": [
    "classifier = classifier.to(args.device)\r\n",
    "dataset.class_weights = dataset.class_weights.to(args.device)\r\n",
    "    \r\n",
    "loss_func = nn.CrossEntropyLoss(dataset.class_weights)\r\n",
    "optimizer = optim.Adam(classifier.parameters(), lr=args.learning_rate)\r\n",
    "scheduler = optim.lr_scheduler.ReduceLROnPlateau(optimizer=optimizer,\r\n",
    "                                           mode='min', factor=0.5,\r\n",
    "                                           patience=1)\r\n",
    "\r\n",
    "train_state = make_train_state(args)\r\n",
    "\r\n",
    "epoch_bar = tqdm(desc='training routine', \r\n",
    "                          total=args.num_epochs,\r\n",
    "                          position=0)\r\n",
    "\r\n",
    "dataset.set_split('train')\r\n",
    "train_bar = tqdm(desc='split=train',\r\n",
    "                          total=dataset.get_num_batches(args.batch_size), \r\n",
    "                          position=1, \r\n",
    "                          leave=True)\r\n",
    "dataset.set_split('val')\r\n",
    "val_bar = tqdm(desc='split=val',\r\n",
    "                        total=dataset.get_num_batches(args.batch_size), \r\n",
    "                        position=1, \r\n",
    "                        leave=True)\r\n",
    "\r\n",
    "try:\r\n",
    "    for epoch_index in range(args.num_epochs):\r\n",
    "        train_state['epoch_index'] = epoch_index\r\n",
    "\r\n",
    "        # Iterate over training dataset\r\n",
    "\r\n",
    "        # setup: batch generator, set loss and acc to 0, set train mode on\r\n",
    "        dataset.set_split('train')\r\n",
    "        batch_generator = generate_batches(dataset, \r\n",
    "                                           batch_size=args.batch_size, \r\n",
    "                                           device=args.device)\r\n",
    "        running_loss = 0.0\r\n",
    "        running_acc = 0.0\r\n",
    "        classifier.train()\r\n",
    "\r\n",
    "        for batch_index, batch_dict in enumerate(batch_generator):\r\n",
    "            # the training routine is these 5 steps:\r\n",
    "\r\n",
    "            # --------------------------------------    \r\n",
    "            # step 1. zero the gradients\r\n",
    "            optimizer.zero_grad()\r\n",
    "\r\n",
    "            # step 2. compute the output\r\n",
    "            y_pred = classifier(x_in=batch_dict['x_data'], \r\n",
    "                                x_lengths=batch_dict['x_length'])\r\n",
    "\r\n",
    "            # step 3. compute the loss\r\n",
    "            loss = loss_func(y_pred, batch_dict['y_target'])\r\n",
    "    \r\n",
    "            running_loss += (loss.item() - running_loss) / (batch_index + 1)\r\n",
    "\r\n",
    "            # step 4. use loss to produce gradients\r\n",
    "            loss.backward()\r\n",
    "\r\n",
    "            # step 5. use optimizer to take gradient step\r\n",
    "            optimizer.step()\r\n",
    "            # -----------------------------------------\r\n",
    "            # compute the accuracy\r\n",
    "            acc_t = compute_accuracy(y_pred, batch_dict['y_target'])\r\n",
    "            running_acc += (acc_t - running_acc) / (batch_index + 1)\r\n",
    "\r\n",
    "            # update bar\r\n",
    "            train_bar.set_postfix(loss=running_loss, acc=running_acc, epoch=epoch_index)\r\n",
    "            train_bar.update()\r\n",
    "\r\n",
    "        train_state['train_loss'].append(running_loss)\r\n",
    "        train_state['train_acc'].append(running_acc)\r\n",
    "\r\n",
    "        # Iterate over val dataset\r\n",
    "\r\n",
    "        # setup: batch generator, set loss and acc to 0; set eval mode on\r\n",
    "\r\n",
    "        dataset.set_split('val')\r\n",
    "        batch_generator = generate_batches(dataset, \r\n",
    "                                           batch_size=args.batch_size, \r\n",
    "                                           device=args.device)\r\n",
    "        running_loss = 0.\r\n",
    "        running_acc = 0.\r\n",
    "        classifier.eval()\r\n",
    "\r\n",
    "        for batch_index, batch_dict in enumerate(batch_generator):\r\n",
    "            # compute the output\r\n",
    "            y_pred = classifier(x_in=batch_dict['x_data'], \r\n",
    "                                x_lengths=batch_dict['x_length'])\r\n",
    "\r\n",
    "            # step 3. compute the loss\r\n",
    "            loss = loss_func(y_pred, batch_dict['y_target'])\r\n",
    "            running_loss += (loss.item() - running_loss) / (batch_index + 1)\r\n",
    "\r\n",
    "            # compute the accuracy\r\n",
    "            acc_t = compute_accuracy(y_pred, batch_dict['y_target'])\r\n",
    "            running_acc += (acc_t - running_acc) / (batch_index + 1)\r\n",
    "            val_bar.set_postfix(loss=running_loss, acc=running_acc, epoch=epoch_index)\r\n",
    "            val_bar.update()\r\n",
    "\r\n",
    "        train_state['val_loss'].append(running_loss)\r\n",
    "        train_state['val_acc'].append(running_acc)\r\n",
    "\r\n",
    "        train_state = update_train_state(args=args, model=classifier, \r\n",
    "                                         train_state=train_state)\r\n",
    "\r\n",
    "        scheduler.step(train_state['val_loss'][-1])\r\n",
    "\r\n",
    "        train_bar.n = 0\r\n",
    "        val_bar.n = 0\r\n",
    "        epoch_bar.update()\r\n",
    "\r\n",
    "        if train_state['stop_early']:\r\n",
    "            break\r\n",
    "            \r\n",
    "except KeyboardInterrupt:\r\n",
    "    print(\"Exiting loop\")"
   ],
   "outputs": [
    {
     "output_type": "display_data",
     "data": {
      "application/vnd.jupyter.widget-view+json": {
       "version_major": 2,
       "version_minor": 0,
       "model_id": "271f5bf60f234cc89dc8557b82e0a225"
      },
      "text/plain": [
       "training routine:   0%|          | 0/100 [00:00<?, ?it/s]"
      ]
     },
     "metadata": {}
    },
    {
     "output_type": "display_data",
     "data": {
      "application/vnd.jupyter.widget-view+json": {
       "version_major": 2,
       "version_minor": 0,
       "model_id": "66018a8d52734cacac1fa21325a9de3d"
      },
      "text/plain": [
       "split=train:   0%|          | 0/120 [00:00<?, ?it/s]"
      ]
     },
     "metadata": {}
    },
    {
     "output_type": "display_data",
     "data": {
      "application/vnd.jupyter.widget-view+json": {
       "version_major": 2,
       "version_minor": 0,
       "model_id": "37b4bb32704e4298a67cc51d4ef6e879"
      },
      "text/plain": [
       "split=val:   0%|          | 0/25 [00:00<?, ?it/s]"
      ]
     },
     "metadata": {}
    }
   ],
   "metadata": {
    "code_folding": []
   }
  },
  {
   "cell_type": "code",
   "execution_count": 24,
   "source": [
    "# compute the loss & accuracy on the test set using the best available model\r\n",
    "\r\n",
    "classifier.load_state_dict(torch.load(train_state['model_filename']))\r\n",
    "\r\n",
    "classifier = classifier.to(args.device)\r\n",
    "dataset.class_weights = dataset.class_weights.to(args.device)\r\n",
    "loss_func = nn.CrossEntropyLoss(dataset.class_weights)\r\n",
    "\r\n",
    "dataset.set_split('test')\r\n",
    "batch_generator = generate_batches(dataset, \r\n",
    "                                   batch_size=args.batch_size, \r\n",
    "                                   device=args.device)\r\n",
    "running_loss = 0.\r\n",
    "running_acc = 0.\r\n",
    "classifier.eval()\r\n",
    "\r\n",
    "for batch_index, batch_dict in enumerate(batch_generator):\r\n",
    "    # compute the output\r\n",
    "    y_pred =  classifier(batch_dict['x_data'],\r\n",
    "                         x_lengths=batch_dict['x_length'])\r\n",
    "    \r\n",
    "    # compute the loss\r\n",
    "    loss = loss_func(y_pred, batch_dict['y_target'])\r\n",
    "    loss_t = loss.item()\r\n",
    "    running_loss += (loss_t - running_loss) / (batch_index + 1)\r\n",
    "\r\n",
    "    # compute the accuracy\r\n",
    "    acc_t = compute_accuracy(y_pred, batch_dict['y_target'])\r\n",
    "    running_acc += (acc_t - running_acc) / (batch_index + 1)\r\n",
    "\r\n",
    "train_state['test_loss'] = running_loss\r\n",
    "train_state['test_acc'] = running_acc"
   ],
   "outputs": [],
   "metadata": {}
  },
  {
   "cell_type": "code",
   "execution_count": 25,
   "source": [
    "print(\"Test loss: {};\".format(train_state['test_loss']))\r\n",
    "print(\"Test Accuracy: {}\".format(train_state['test_acc']))"
   ],
   "outputs": [
    {
     "output_type": "stream",
     "name": "stdout",
     "text": [
      "Test loss: 1.8524044704437257;\n",
      "Test Accuracy: 41.06249999999999\n"
     ]
    }
   ],
   "metadata": {}
  },
  {
   "cell_type": "markdown",
   "source": [
    "### Inference"
   ],
   "metadata": {}
  },
  {
   "cell_type": "code",
   "execution_count": 26,
   "source": [
    "def predict_nationality(surname, classifier, vectorizer):\r\n",
    "    vectorized_surname, vec_length = vectorizer.vectorize(surname)\r\n",
    "    vectorized_surname = torch.tensor(vectorized_surname).unsqueeze(dim=0)\r\n",
    "    vec_length = torch.tensor([vec_length], dtype=torch.int64)\r\n",
    "    \r\n",
    "    result = classifier(vectorized_surname, vec_length, apply_softmax=True)\r\n",
    "    probability_values, indices = result.max(dim=1)\r\n",
    "    \r\n",
    "    index = indices.item()\r\n",
    "    prob_value = probability_values.item()\r\n",
    "\r\n",
    "    predicted_nationality = vectorizer.nationality_vocab.lookup_index(index)\r\n",
    "\r\n",
    "    return {'nationality': predicted_nationality, 'probability': prob_value, 'surname': surname}"
   ],
   "outputs": [],
   "metadata": {}
  },
  {
   "cell_type": "code",
   "execution_count": 27,
   "source": [
    "# surname = input(\"Enter a surname: \")\r\n",
    "classifier = classifier.to(\"cpu\")\r\n",
    "for surname in ['McMahan', 'Nakamoto', 'Wan', 'Cho']:\r\n",
    "    print(predict_nationality(surname, classifier, vectorizer))"
   ],
   "outputs": [
    {
     "output_type": "stream",
     "name": "stdout",
     "text": [
      "{'nationality': 'Irish', 'probability': 0.4722009301185608, 'surname': 'McMahan'}\n",
      "{'nationality': 'Japanese', 'probability': 0.7476573586463928, 'surname': 'Nakamoto'}\n",
      "{'nationality': 'Chinese', 'probability': 0.6742308139801025, 'surname': 'Wan'}\n",
      "{'nationality': 'Korean', 'probability': 0.3817710876464844, 'surname': 'Cho'}\n"
     ]
    }
   ],
   "metadata": {}
  },
  {
   "cell_type": "markdown",
   "source": [
    ":::{admonition} Your Turn\r\n",
    "\r\n",
    "- Play with the hyperparameters to get a sense of what affects performance and by how much, and to tabulate the results. \r\n",
    "- The model implemented in here is general and not restricted to characters. The embedding layer in the model can map any discrete item in a sequence of discrete items. Consider reuse the code here for sentence or document classification. \r\n",
    "\r\n",
    "\r\n"
   ],
   "metadata": {}
  }
 ],
 "metadata": {
  "kernelspec": {
   "name": "python3",
   "display_name": "Python 3.8.10 64-bit ('cits4012': conda)"
  },
  "language_info": {
   "codemirror_mode": {
    "name": "ipython",
    "version": 3
   },
   "file_extension": ".py",
   "mimetype": "text/x-python",
   "name": "python",
   "nbconvert_exporter": "python",
   "pygments_lexer": "ipython3",
   "version": "3.8.10"
  },
  "toc": {
   "colors": {
    "hover_highlight": "#DAA520",
    "running_highlight": "#FF0000",
    "selected_highlight": "#FFD700"
   },
   "moveMenuLeft": true,
   "nav_menu": {
    "height": "120px",
    "width": "252px"
   },
   "navigate_menu": true,
   "number_sections": true,
   "sideBar": true,
   "threshold": "5",
   "toc_cell": false,
   "toc_section_display": "block",
   "toc_window_display": false
  },
  "interpreter": {
   "hash": "d990147e05fc0cc60dd3871899a6233eb6a5324c1885ded43d013dc915f7e535"
  }
 },
 "nbformat": 4,
 "nbformat_minor": 2
}